@@ -55,8 +55,10 @@ typedef struct fd_queue_elem_t fd_queue_elem_t;
 
 struct fd_queue_t {
     fd_queue_elem_t    *data;
-    int                 nelts;
-    int                 bounds;
+    unsigned int       nelts;
+    unsigned int       bounds;
+    unsigned int       in;
+    unsigned int       out;
     apr_thread_mutex_t *one_big_mutex;
     apr_thread_cond_t  *not_empty;
     int                 terminated;
